using System;
using System.IO.Ports;
using System.Threading;
using System.Threading.Tasks;
using NModbus;
using NModbus.Serial;
using UnityEngine;

namespace Assets.Scripts.ModBus
{
    public class ModBusRtu
    {
        private static ModBusRtu m_Instance;
        public static ModBusRtu Instance
        {
            get
            {
                if (m_Instance == null)
                    m_Instance = new ModBusRtu();
                return m_Instance;
            }
        }
        
        private string masterPortName = "COM2";
        private string slavePortName = "COM3";
        private byte slaveAddress = 0x01;
        private ushort startAddress = 0;
        private ushort numberOfPoints = 10;

        private CancellationToken token;
        private CancellationTokenSource cts;
        
        public void Init()
        {
            if (cts != null)
            {
                cts.Cancel();
                cts = null;
            }
            cts = new CancellationTokenSource();
            token = cts.Token;
        }

        public void Dispose()
        {
            cts.Cancel();
        }
        
        public ushort[] ModbusSerialRtuMasterReadHoldingRegisters()
        {
            // configure serial port
            using SerialPort masterPort = new SerialPort(masterPortName);
            masterPort.BaudRate = 9600;
            masterPort.DataBits = 8;
            masterPort.Parity = Parity.None;
            masterPort.StopBits = StopBits.One;
            masterPort.Open();

            var factory = new ModbusFactory();
            IModbusSerialMaster master = factory.CreateRtuMaster(masterPort);

            // read register
            ushort[] registers = master.ReadHoldingRegisters(slaveAddress, startAddress, numberOfPoints);
            return registers;
        }

        public ushort[] ModbusSerialRtuMasterReadInputRegisters()
        {
            // configure serial port
            using SerialPort masterPort = new SerialPort(masterPortName);
            masterPort.BaudRate = 9600;
            masterPort.DataBits = 8;
            masterPort.Parity = Parity.None;
            masterPort.StopBits = StopBits.One;
            masterPort.Open();

            var factory = new ModbusFactory();
            IModbusSerialMaster master = factory.CreateRtuMaster(masterPort);

            // read register
            ushort[] registers = master.ReadInputRegisters(slaveAddress, startAddress, numberOfPoints);
            return registers;
        }

        public void ModbusSerialRtuMasterWriteMultipleRegisters()
        {
            // configure serial port
            using SerialPort port = new SerialPort(masterPortName);
            port.BaudRate = 9600;
            port.DataBits = 8;
            port.Parity = Parity.None;
            port.StopBits = StopBits.One;
            port.Open();

            var factory = new ModbusFactory();
            IModbusMaster master = factory.CreateRtuMaster(port);
                
            ushort[] data = new ushort[] { 10, 20, 30, 0, 0 };

            // write register
            master.WriteMultipleRegisters(slaveAddress, startAddress, data);
        }

        public void StartModbusSerialRtuSlave()
        {
            StartModbusSerialRtuSlaveWithCustomStoreAsync(token);
        }

        public async void StartModbusSerialRtuSlaveWithCustomStoreAsync(CancellationToken ct)
        {
            if (ct.IsCancellationRequested) return;
            // configure serial port
            using SerialPort slavePort = new SerialPort(slavePortName);
            slavePort.BaudRate = 9600;
            slavePort.DataBits = 8;
            slavePort.Parity = Parity.None;
            slavePort.StopBits = StopBits.One;
            slavePort.Open();

            var factory = new ModbusFactory();
            var slaveNetwork = factory.CreateRtuSlaveNetwork(slavePort);

            slavePort.ReadTimeout = 50;
            slavePort.WriteTimeout = 1000;

            var dataStore = new SlaveStorage();

            dataStore.CoilDiscretes.StorageOperationOccurred += (sender, args) =>
            {
                Debug.Log(GetSlaveMsg("CoilDiscretes", args));
            };
            dataStore.CoilInputs.StorageOperationOccurred += (sender, args) =>
            {
                Debug.Log(GetSlaveMsg("CoilInputs", args));

            };
            dataStore.InputRegisters.StorageOperationOccurred += (sender, args) =>
            {
                Debug.Log(GetSlaveMsg("InputRegisters", args));

            };
            dataStore.HoldingRegisters.StorageOperationOccurred += (sender, args) =>
            {
                Debug.Log(GetSlaveMsg("HoldingRegisters", args));
            };

            IModbusSlave slave = factory.CreateSlave(slaveAddress, dataStore);

            slaveNetwork.AddSlave(slave);

            // listen async
            await Task.Run(() => { slaveNetwork.ListenAsync(ct); }, ct);
        }

        private string GetSlaveMsg(string type, StorageEventArgs<bool> args)
        {
            string value = type + $":{args.Operation} StartingAddress:{args.StartingAddress} ";
            for (int i = 0; i < args.Points.Length; i++)
            {
                value += $"Point{i + args.StartingAddress}:{args.Points[i]} ";
            }
            return value;
        }

        private string GetSlaveMsg(string type, StorageEventArgs<ushort> args)
        {
            string value = type + $":{args.Operation} StartingAddress:{args.StartingAddress} ";
            for (int i = 0; i < args.Points.Length; i++)
            {
                value += $"Point{i + args.StartingAddress}:{args.Points[i]} ";
            }
            return value;
        }

        public void StartModbusBand()
        {
            // configure serial port
            using SerialPort port = new SerialPort(masterPortName);
            port.BaudRate = 9600;
            port.DataBits = 8;
            port.Parity = Parity.None;
            port.StopBits = StopBits.One;
            port.Open();

            var factory = new ModbusFactory();
            IModbusMaster master = factory.CreateRtuMaster(port);
        }
    }
}