﻿using System;
using System.Collections.Generic;
using NModbus;

namespace Assets.Scripts.ModBus
{
    public class SlaveStorage : ISlaveDataStore
    {
        private readonly SparsePointSource<bool> m_CoilDiscretes;
        private readonly SparsePointSource<bool> m_CoilInputs;
        private readonly SparsePointSource<ushort> m_HoldingRegisters;
        private readonly SparsePointSource<ushort> m_InputRegisters;

        public SlaveStorage()
        {
            m_CoilDiscretes = new SparsePointSource<bool>();
            m_CoilInputs = new SparsePointSource<bool>();
            m_HoldingRegisters = new SparsePointSource<ushort>();
            m_InputRegisters = new SparsePointSource<ushort>();
        }

        public SparsePointSource<bool> CoilDiscretes => m_CoilDiscretes;

        public SparsePointSource<bool> CoilInputs => m_CoilInputs;

        public SparsePointSource<ushort> HoldingRegisters => m_HoldingRegisters;
        
        public SparsePointSource<ushort> InputRegisters => m_InputRegisters;

        IPointSource<bool> ISlaveDataStore.CoilDiscretes => m_CoilDiscretes;

        IPointSource<bool> ISlaveDataStore.CoilInputs => m_CoilInputs;

        IPointSource<ushort> ISlaveDataStore.HoldingRegisters => m_HoldingRegisters;

        IPointSource<ushort> ISlaveDataStore.InputRegisters => m_InputRegisters;
    }

    /// <summary>
    /// Sparse storage for points.
    /// </summary>
    public class SparsePointSource<TPoint> : IPointSource<TPoint>
    {
        private readonly Dictionary<ushort, TPoint> _values = new Dictionary<ushort, TPoint>();

        public event EventHandler<StorageEventArgs<TPoint>> StorageOperationOccurred;

        /// <summary>
        /// Gets or sets the value of an individual point wih tout 
        /// </summary>
        /// <param name="registerIndex"></param>
        /// <returns></returns>
        public TPoint this[ushort registerIndex]
        {
            get
            {
                if (_values.TryGetValue(registerIndex, out var value))
                    return value;

                return default(TPoint);
            }
            set { _values[registerIndex] = value; }
        }

        public TPoint[] ReadPoints(ushort startAddress, ushort numberOfPoints)
        {
            var points = new TPoint[numberOfPoints];

            for (ushort index = 0; index < numberOfPoints; index++)
            {
                points[index] = this[(ushort) (index + startAddress)];
            }

            StorageOperationOccurred?.Invoke(this,
                new StorageEventArgs<TPoint>(PointOperation.Read, startAddress, points));

            return points;
        }

        public void WritePoints(ushort startAddress, TPoint[] points)
        {
            for (ushort index = 0; index < points.Length; index++)
            {
                this[(ushort) (index + startAddress)] = points[index];
            }

            StorageOperationOccurred?.Invoke(this,
                new StorageEventArgs<TPoint>(PointOperation.Write, startAddress, points));
        }
    }

    public class StorageEventArgs<TPoint> : EventArgs
    {
        public StorageEventArgs(PointOperation pointOperation, ushort startingAddress, TPoint[] points)
        {
            Operation = pointOperation;
            StartingAddress = startingAddress;
            Points = points;
        }

        public ushort StartingAddress { get; }

        public TPoint[] Points { get; }

        public PointOperation Operation { get; }
    }

    public enum PointOperation
    {
        Read,
        Write
    }
}